from typing import Tuple
import pygame
import pygame.display
import pygame.event
import pygame._sdl2

from moderngl_window.context.base import BaseWindow
from moderngl_window.context.pygame2.keys import Keys


class Window(BaseWindow):
    """
    Basic window implementation using pygame2.
    """

    #: Name of the window
    name = "pygame2"
    #: pygame specific key constants
    keys = Keys

    _mouse_button_map = {
        1: 1,
        3: 2,
        2: 3,
    }

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        pygame.display.init()

        pygame.display.gl_set_attribute(
            pygame.GL_CONTEXT_MAJOR_VERSION, self.gl_version[0]
        )
        pygame.display.gl_set_attribute(
            pygame.GL_CONTEXT_MINOR_VERSION, self.gl_version[1]
        )
        pygame.display.gl_set_attribute(
            pygame.GL_CONTEXT_PROFILE_MASK, pygame.GL_CONTEXT_PROFILE_CORE
        )
        pygame.display.gl_set_attribute(pygame.GL_CONTEXT_FORWARD_COMPATIBLE_FLAG, 1)
        pygame.display.gl_set_attribute(pygame.GL_DOUBLEBUFFER, 1)
        pygame.display.gl_set_attribute(pygame.GL_DEPTH_SIZE, 24)
        pygame.display.gl_set_attribute(pygame.GL_STENCIL_SIZE, 8)

        if self.samples > 1:
            pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLEBUFFERS, 1)
            pygame.display.gl_set_attribute(pygame.GL_MULTISAMPLESAMPLES, self.samples)

        self._depth = 24
        self._flags = pygame.OPENGL | pygame.DOUBLEBUF

        if self.resizable:
            self._flags |= pygame.RESIZABLE

        if not self._visible:
            self._flags |= pygame.HIDDEN

        self._set_mode()
        self.title = self._title
        self.cursor = self._cursor
        # Get the reference for the internal sdl2 window
        # Makes us able to control window position and other properties.
        self._sdl_window = pygame._sdl2.video.Window.from_display_module()

        if self.fullscreen:
            self._set_fullscreen(True)

        self.init_mgl_context()
        self.set_default_viewport()

    def _set_mode(self):
        self._surface = pygame.display.set_mode(
            size=(self._width, self._height), flags=self._flags, depth=self._depth, vsync=self._vsync
        )

    def _set_fullscreen(self, value: bool) -> None:
        if value:
            self._sdl_window.set_fullscreen(True)
        else:
            self._sdl_window.set_windowed()

    def _set_vsync(self, value: bool) -> None:
        self._vsync = value
        self._set_mode()

    @property
    def size(self) -> Tuple[int, int]:
        """Tuple[int, int]: current window size.

        This property also support assignment::

            # Resize the window to 1000 x 1000
            window.size = 1000, 1000
        """
        return self._width, self._height

    @size.setter
    def size(self, value: Tuple[int, int]):
        self._width, self._height = value
        self._set_mode()
        self.resize(value[0], value[1])

    @property
    def position(self) -> Tuple[int, int]:
        """Tuple[int, int]: The current window position.

        This property can also be set to move the window::

            # Move window to 100, 100
            window.position = 100, 100
        """
        return self._sdl_window.position

    @position.setter
    def position(self, value: Tuple[int, int]):
        self._sdl_window.position = value

    @property
    def visible(self) -> bool:
        """bool: Is the window visible?

        This property can also be set::

            # Hide or show the window
            window.visible = False
        """
        return self._visible

    @visible.setter
    def visible(self, value: bool):
        self._visible = value
        if value:
            self._sdl_window.show()
        else:
            self._sdl_window.hide()

    @property
    def cursor(self) -> bool:
        """bool: Should the mouse cursor be visible inside the window?

        This property can also be assigned to::

            # Disable cursor
            window.cursor = False
        """
        return self._cursor

    @cursor.setter
    def cursor(self, value: bool):
        pygame.mouse.set_visible(value)
        self._cursor = value

    @property
    def mouse_exclusivity(self) -> bool:
        """bool: If mouse exclusivity is enabled.

        When you enable mouse-exclusive mode, the mouse cursor is no longer
        available. It is not merely hidden – no amount of mouse movement
        will make it leave your application. This is for example useful
        when you don't want the mouse leaving the screen when rotating
        a 3d scene.

        This property can also be set::

            window.mouse_exclusivity = True
        """
        return self._mouse_exclusivity

    @mouse_exclusivity.setter
    def mouse_exclusivity(self, value: bool):
        if self._cursor:
            self.cursor = False

        pygame.event.set_grab(value)
        self._mouse_exclusivity = value

    @property
    def title(self) -> str:
        """str: Window title.

        This property can also be set::

            window.title = "New Title"
        """
        return self._title

    @title.setter
    def title(self, value: str):
        pygame.display.set_caption(value)
        self._title = value

    def swap_buffers(self) -> None:
        """Swap buffers, set viewport, trigger events and increment frame counter"""
        pygame.display.flip()
        self.set_default_viewport()
        self.process_events()
        self._frames += 1

    def _set_icon(self, icon_path: str) -> None:
        icon = pygame.image.load(icon_path)
        pygame.display.set_icon(icon)

    def resize(self, width, height) -> None:
        """Resize callback

        Args:
            width: New window width
            height: New window height
        """
        self._width = width
        self._height = height
        self._buffer_width, self._buffer_height = self._width, self._height
        self.set_default_viewport()

        super().resize(self._buffer_width, self._buffer_height)

    def close(self):
        """Close the window"""
        super().close()
        self._close_func()

    def _handle_mods(self) -> None:
        """Update key mods"""
        mods = pygame.key.get_mods()
        self._modifiers.shift = mods & pygame.KMOD_SHIFT
        self._modifiers.ctrl = mods & pygame.KMOD_CTRL
        self._modifiers.alt = mods & pygame.KMOD_ALT

    def process_events(self) -> None:
        """Handle all queued events in pygame2 dispatching events to standard methods"""

        for event in pygame.event.get():
            if event.type == pygame.MOUSEMOTION:
                self._handle_mods()
                if self.mouse_states.any:
                    self._mouse_drag_event_func(
                        event.pos[0], event.pos[1], event.rel[0], event.rel[1],
                    )
                else:
                    self._mouse_position_event_func(
                        event.pos[0], event.pos[1], event.rel[0], event.rel[1],
                    )

            elif event.type == pygame.MOUSEBUTTONDOWN:
                self._handle_mods()
                button = self._mouse_button_map.get(event.button, None)
                if button is not None:
                    self._handle_mouse_button_state_change(button, True)
                    self._mouse_press_event_func(
                        event.pos[0], event.pos[1], button,
                    )

            elif event.type == pygame.MOUSEBUTTONUP:
                self._handle_mods()
                button = self._mouse_button_map.get(event.button, None)
                if button is not None:
                    self._handle_mouse_button_state_change(button, False)
                    self._mouse_release_event_func(
                        event.pos[0], event.pos[1], button,
                    )

            elif event.type in [pygame.KEYDOWN, pygame.KEYUP]:
                self._handle_mods()

                if self._exit_key is not None and event.key == self._exit_key:
                    self.close()

                # Pygame can't do fullscreen yet, but this would toggle it.
                if event.type == pygame.KEYUP and self._fs_key is not None and event.key == self._fs_key:
                    self.fullscreen = not self.fullscreen

                if event.type == pygame.KEYDOWN:
                    self._key_pressed_map[event.key] = True
                elif event.type == pygame.KEYUP:
                    self._key_pressed_map[event.key] = False

                self._key_event_func(event.key, event.type, self._modifiers)

            elif event.type == pygame.TEXTINPUT:
                self._handle_mods()
                self._unicode_char_entered_func(event.text)

            elif event.type == pygame.MOUSEWHEEL:
                self._handle_mods()
                self._mouse_scroll_event_func(float(event.x), float(event.y))

            elif event.type == pygame.QUIT:
                self.close()

            elif event.type == pygame.VIDEORESIZE:
                self.resize(event.size[0], event.size[1])

            elif event.type == pygame.ACTIVEEVENT:

                # # We might support these in the future
                # Mouse cursor state
                # if event.state == 0:
                #     if event.gain:
                #         print("Mouse enters viewport")
                #     else:
                #         print("Mouse leaves viewport")

                # Window focus state
                # if event.state == 1:
                #     if event.gain:
                #         print("Window gained focus")
                #     else:
                #         print("Window lost focus")

                # Window iconify state
                if getattr(event, 'state', None) == 2:
                    if event.gain:
                        self._iconify_func(False)
                    else:
                        self._iconify_func(True)

            # This is also a problem on linux, but is too disruptive during resize events
            # elif event.type == pygame.VIDEOEXPOSE:
            #     # On OS X we only get VIDEOEXPOSE when restoring the windoe
            #     self._iconify_func(False)

            elif event.type == pygame.USEREVENT:
                self._on_generic_event_func(event)

    def destroy(self) -> None:
        """Gracefully close the window"""
        pygame.quit()



from pathlib import Path
from pyrr import Matrix44

import moderngl

import moderngl_window
from moderngl_window import geometry
from base import CameraWindow


class TextureArrayExample(CameraWindow):
    """
    Cycles different texture layers in an array texture
    rendered on a cube.
    """
    title = "Texture Array"
    resource_dir = (Path(__file__).parent / 'resources').resolve()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True
        self.num_layers = 10
        self.cube = geometry.cube(size=(2, 2, 2))
        self.texture = self.load_texture_array(
            'textures/array.png', layers=self.num_layers, mipmap=True, anisotrpy=8.0)
        self.prog = self.load_program('programs/cube_texture_array.glsl')
        self.prog['texture0'].value = 0
        self.prog['num_layers'].value = 10

    def render(self, time: float, frametime: float):
        self.ctx.enable_only(moderngl.CULL_FACE | moderngl.DEPTH_TEST)

        rotation = Matrix44.from_eulers((time, time, time), dtype='f4')
        translation = Matrix44.from_translation((0.0, 0.0, -3.5), dtype='f4')
        modelview = translation * rotation

        self.prog['m_proj'].write(self.camera.projection.matrix)
        self.prog['m_model'].write(modelview)
        self.prog['m_camera'].write(self.camera.matrix)
        self.prog['time'].value = time

        self.texture.use(location=0)
        self.cube.render(self.prog)


if __name__ == '__main__':
    moderngl_window.run_window_config(TextureArrayExample)



from moderngl_window.loaders.base import BaseLoader
from moderngl_window.finders import texture
from pathlib import Path


class IconLoader(BaseLoader):
    kind = "icon"

    def __init__(self, meta):
        super().__init__(meta)

    def find_icon(self):
        """Find resource using texture finders.

        This is mainly a shortcut method to simplify the task.

        Args:
            path: Path to resource
        """
        abs_path = self._find(Path(self.meta.path), texture.get_finders())
        if abs_path is None:
            raise ValueError(
                "Could not find the icon specified. {}".format(self.meta.path)
            )
        return abs_path



"""
Notes from moderngl:

The vao_content is a list of 3-tuples (buffer, format, attribs)
the format can have an empty or '/v', '/i', '/r' ending.
'/v' attributes are the default
'/i` attributes are per instance attributes
'/r' attributes are per render (like a uniform)
Example:
    vao_content = [
        (self.position_vertex_buffer, '2f', 'in_vert'),
        (self.color_buffer, '3f', 'in_color'),
        (self.pos_scale_buffer, '2f 1f/i', 'in_pos', 'in_scale'),
    ]
"""
import re
from functools import lru_cache
from typing import List

VALID_DIVISORS = ["v", "i", "r"]


class BufferFormat:
    def __init__(
        self,
        format_string: str,
        components: int,
        bytes_per_component: int,
        per_instance=False,
    ):
        """
        Args:
            format_string (str): moderngl format string
            components (int): components
            byte_size (int): bytes per component
            per_instance (bool): Instanced attribute
        """
        self.format = format_string
        self.components = components
        self.bytes_per_component = bytes_per_component
        self.per_instance = per_instance

    @property
    def bytes_total(self) -> int:
        """int: total byte size if this type"""
        return self.components * self.bytes_per_component

    def pad_str(self) -> str:
        """Padding string used my moderngl in interleaved buffers"""
        return "{}x{}".format(self.components, self.bytes_per_component)

    def __str__(self) -> str:
        return "<BufferFormat {} components={} bytes_per_component={}>".format(
            self.format, self.components, self.bytes_per_component
        )

    def __repr__(self) -> str:
        return str(self)


@lru_cache(maxsize=500)
def attribute_format(attr_format: str) -> BufferFormat:
    """Look up info about an attribute format.

    Translate the format into a BufferFormat instance
    containing things like byte size and components

    Args:
        buffer_format (str): Format of an attribute
    Returns:
        BufferFormat instance
    """
    if not attr_format:
        raise ValueError("Cannot resolve buffer format: '{}'".format(attr_format))

    parts = attr_format.split("/")

    # Parse out divisor if present
    # Examples: 3f/i, 3f/v
    fmt = parts[0]
    divisor = ""
    if len(parts) > 1:
        divisor = parts[1]
        if divisor not in VALID_DIVISORS:
            raise ValueError(
                "Invalid attribute divisor '{}' in '{}'".format(divisor, buffer_format)
            )

    # Parse out out component count and actual format
    parts = re.split(r"([fiudn])", fmt)
    components = 1
    if parts[0].isalnum():
        components = int(parts[0])
        bformat = fmt[len(parts[0]):]
    else:
        bformat = fmt

    # Construct specific buffer format
    fmt_info = buffer_format(bformat)
    return BufferFormat(
        "{}{}{}".format(components, bformat, "/{}".format(divisor) if divisor else ""),
        components,
        fmt_info.bytes_per_component,
        per_instance=divisor == "i",
    )


def parse_attribute_formats(frmt: str) -> List[BufferFormat]:
    return [attribute_format(attr) for attr in frmt.split()]


def buffer_format(frmt: str) -> BufferFormat:
    """Look up info about a buffer format type

    Args:
        frmt (str): format string such as 'f', 'i' and 'u'
    Returns:
        BufferFormat instance
    """
    try:
        return BUFFER_FORMATS[frmt]
    except KeyError:
        raise ValueError(
            "Buffer format '{}' unknown. Valid formats: {}".format(
                frmt, BUFFER_FORMATS.keys()
            )
        )


BUFFER_FORMATS = {
    # Floats
    "f": BufferFormat("f", 1, 4),
    "f1": BufferFormat("f1", 1, 1),
    "f2": BufferFormat("f2", 1, 2),
    "f4": BufferFormat("f4", 1, 4),
    "f8": BufferFormat("f8", 1, 8),
    # Unsigned Integers
    "u": BufferFormat("u", 1, 4),
    "u1": BufferFormat("u1", 1, 1),
    "u2": BufferFormat("u2", 1, 2),
    "u4": BufferFormat("u4", 1, 4),
    # Signed Integer
    "i": BufferFormat("i", 1, 4),
    "i1": BufferFormat("i1", 1, 1),
    "i2": BufferFormat("i2", 1, 2),
    "i4": BufferFormat("i4", 1, 4),
    # Normalized float
    "nf": BufferFormat("nf", 1, 4),
    "nf1": BufferFormat("nf1", 1, 1),
    "nf2": BufferFormat("nf2", 1, 2),
    "nf4": BufferFormat("nf4", 1, 4),
    # Normalized uint
    "nu": BufferFormat("nu", 1, 4),
    "nu1": BufferFormat("nu1", 1, 1),
    "nu2": BufferFormat("nu2", 1, 2),
    "nu4": BufferFormat("nu4", 1, 4),
    # Normalized int
    "ni": BufferFormat("ni", 1, 4),
    "ni1": BufferFormat("ni1", 1, 1),
    "ni2": BufferFormat("ni2", 1, 2),
    "ni4": BufferFormat("ni4", 1, 4),
}



import logging
import os

import numpy

import pywavefront
from pywavefront import cache
from pywavefront.obj import ObjParser

import moderngl
from moderngl_window.loaders.base import BaseLoader
from moderngl_window.opengl.vao import VAO
from moderngl_window import resources
from moderngl_window.resources.decorators import texture_dirs
from moderngl_window.meta import SceneDescription, TextureDescription
from moderngl_window.scene import Material, MaterialTexture, Mesh, Node, Scene
from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.geometry.attributes import AttributeNames


logger = logging.getLogger(__name__)


def translate_buffer_format(vertex_format: str, attr_names: AttributeNames):
    """Translate the buffer format"""
    buffer_format = []
    attributes = []
    mesh_attributes = []

    if "T2F" in vertex_format:
        buffer_format.append("2f")
        attributes.append(attr_names.TEXCOORD_0)
        mesh_attributes.append(("TEXCOORD_0", attr_names.TEXCOORD_0, 2))

    if "C3F" in vertex_format:
        buffer_format.append("3f")
        attributes.append(attr_names.COLOR_0)
        mesh_attributes.append(("COLOR_0", attr_names.COLOR_0, 3))

    if "N3F" in vertex_format:
        buffer_format.append("3f")
        attributes.append(attr_names.NORMAL)
        mesh_attributes.append(("NORMAL", attr_names.NORMAL, 3))

    buffer_format.append("3f")
    attributes.append(attr_names.POSITION)
    mesh_attributes.append(("POSITION", attr_names.POSITION, 3))

    return " ".join(buffer_format), attributes, mesh_attributes


class VAOCacheLoader(cache.CacheLoader):
    """Load geometry data directly into vaos"""

    attr_names = None

    def load_vertex_buffer(self, fd, material, length):
        buffer_format, attributes, mesh_attributes = translate_buffer_format(
            material.vertex_format, self.attr_names
        )

        vao = VAO(material.name, mode=moderngl.TRIANGLES)
        vao.buffer(fd.read(length), buffer_format, attributes)

        setattr(material, "vao", vao)
        setattr(material, "buffer_format", buffer_format)
        setattr(material, "attributes", attributes)
        setattr(material, "mesh_attributes", mesh_attributes)


ObjParser.cache_loader_cls = VAOCacheLoader


class Loader(BaseLoader):
    """Loade wavefront/obj files"""

    kind = "wavefront"
    file_extensions = [
        [".obj"],
        [".obj", ".gz"],
        [".bin"],
    ]

    def __init__(self, meta: SceneDescription):
        super().__init__(meta)

    def load(self):
        """Loads a wavefront/obj file including materials and textures

        Returns:
            Scene: The Scene instance
        """
        path = self.find_scene(self.meta.path)
        logger.info("loading %s", path)

        if not path:
            raise ImproperlyConfigured("Scene '{}' not found".format(self.meta.path))

        if path.suffix == ".bin":
            path = path.parent / path.stem

        VAOCacheLoader.attr_names = self.meta.attr_names

        data = pywavefront.Wavefront(
            str(path), create_materials=True, cache=self.meta.cache
        )
        scene = Scene(self.meta.resolved_path)
        texture_cache = {}

        for _, mat in data.materials.items():
            mesh = Mesh(mat.name)

            # Traditional loader
            if mat.vertices:
                buffer_format, attributes, mesh_attributes = translate_buffer_format(
                    mat.vertex_format, self.meta.attr_names
                )
                vbo = numpy.array(mat.vertices, dtype="f4")

                vao = VAO(mat.name, mode=moderngl.TRIANGLES)
                vao.buffer(vbo, buffer_format, attributes)
                mesh.vao = vao

                for attrs in mesh_attributes:
                    mesh.add_attribute(*attrs)

            # Binary cache loader
            elif hasattr(mat, "vao"):
                mesh = Mesh(mat.name)
                mesh.vao = mat.vao
                for attrs in mat.mesh_attributes:
                    mesh.add_attribute(*attrs)
            else:
                # Empty
                continue

            scene.meshes.append(mesh)

            mesh.material = Material(mat.name)
            scene.materials.append(mesh.material)
            mesh.material.color = mat.diffuse

            if mat.texture:
                # A texture can be referenced multiple times, so we need to cache loaded ones
                texture = texture_cache.get(mat.texture.path)
                if not texture:
                    # HACK: pywavefront only give us an absolute path
                    rel_path = os.path.relpath(mat.texture.find(), str(path.parent))
                    logger.info("Loading: %s", rel_path)
                    with texture_dirs([path.parent]):
                        texture = resources.textures.load(
                            TextureDescription(
                                label=rel_path,
                                path=rel_path,
                                mipmap=True,
                                anisotropy=16.0,
                            )
                        )
                    texture_cache[rel_path] = texture

                mesh.material.mat_texture = MaterialTexture(
                    texture=texture, sampler=None,
                )

            node = Node(mesh=mesh)
            scene.root_nodes.append(node)

        # Not supported yet for obj
        # self.calc_scene_bbox()
        scene.prepare()

        return scene



import numpy
from pyrr import matrix44

from pathlib import Path

import moderngl
from moderngl_window.opengl.vao import VAO
from moderngl_window import resources
from moderngl_window.meta import (
    DataDescription,
    TextureDescription,
    ProgramDescription,
)

from .base import BaseText, FontMeta

resources.register_dir(Path(__file__).parent.resolve())


class TextWriter2D(BaseText):
    """Simple monspaced bitmapped text renderer"""

    def __init__(self):
        super().__init__()

        meta = FontMeta(resources.data.load(
            DataDescription(path="bitmapped/text/meta.json")
        ))
        self._texture = resources.textures.load(
            TextureDescription(
                path="bitmapped/textures/VeraMono.png",
                kind="array",
                mipmap=True,
                layers=meta.characters,
            )
        )
        self._program = resources.programs.load(
            ProgramDescription(path="bitmapped/programs/text_2d.glsl")
        )

        self._init(meta)

        self._string_buffer = self.ctx.buffer(reserve=1024 * 4)
        self._string_buffer.clear(chunk=b'\32')
        pos = self.ctx.buffer(data=bytes([0] * 4 * 3))

        self._vao = VAO("textwriter", mode=moderngl.POINTS)
        self._vao.buffer(pos, '3f', 'in_position')
        self._vao.buffer(self._string_buffer, '1u/i', 'in_char_id')

        self._text: str = None

    @property
    def text(self) -> str:
        return self._text

    @text.setter
    def text(self, value: str):
        self._text = value
        self._string_buffer.orphan(size=len(value) * 4)
        self._string_buffer.clear(chunk=b'\32')
        self._write(value)

    def _write(self, text: str):
        self._string_buffer.clear(chunk=b'\32')

        self._string_buffer.write(
            numpy.fromiter(
                self._translate_string(text),
                dtype=numpy.uint32,
            )
        )

    def draw(self, pos, length=-1, size=24.0):
        # Calculate ortho projection based on viewport
        vp = self.ctx.fbo.viewport
        w, h = vp[2], vp[3]
        projection = matrix44.create_orthogonal_projection_matrix(
            0,  # left
            w,  # right
            0,  # bottom
            h,  # top
            1.0,  # near
            -1.0,  # far
            dtype=numpy.float32,
        )

        self._texture.use(location=0)
        self._program["m_proj"].write(projection)
        self._program["text_pos"].value = pos
        self._program["font_texture"].value = 0
        self._program["char_size"].value = self._meta.char_aspect_wh * size, size

        self._vao.render(self._program, instances=len(self._text))



from pathlib import Path
from pyrr import Matrix44
import numpy

import moderngl
import moderngl_window
from moderngl_window import geometry
from base import CameraWindow


class LinesDemo(CameraWindow):
    """Rendering thick lines with geometry shader

    Example is basic and incomplete, but shows how
    one could use the geometry shader to create thick lines.
    """
    gl_version = (3, 3)
    title = "Thick Lines"
    resource_dir = (Path(__file__) / '../resources').absolute()

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.wnd.mouse_exclusivity = True

        self.prog = self.load_program('programs/lines/lines.glsl')
        self.prog['color'].value = (1.0, 1.0, 1.0, 1.0)
        self.prog['m_model'].write(Matrix44.from_translation((0.0, 0.0, -3.5), dtype='f4'))

        N = 10
        # Create lines geometry
        def gen_lines():
            for i in range(N):
                # A
                yield -1.0
                yield 1.0 - i * 2.0 / N
                yield 0.0
                # b
                yield 1.0
                yield 1.0 - i * 2.0 / N
                yield 0.0

        buffer = self.ctx.buffer(numpy.fromiter(gen_lines(), dtype='f4', count=N * 6).tobytes())
        self.lines = self.ctx.vertex_array(
            self.prog,
            [
                (buffer, '3f', 'in_position'),
            ],
        )

    def render(self, time, frametime):
        # self.ctx.enable_only(moderngl.DEPTH_TEST)

        self.prog['m_proj'].write(self.camera.projection.matrix)
        self.prog['m_cam'].write(self.camera.matrix)
        self.lines.render(mode=moderngl.LINES)


if __name__ == '__main__':
    moderngl_window.run_window_config(LinesDemo)



from contextlib import contextmanager
from pathlib import Path
from typing import Union

from moderngl_window.conf import settings
from moderngl_window.exceptions import ImproperlyConfigured

from moderngl_window.resources.programs import programs  # noqa
from moderngl_window.resources.textures import textures  # noqa

# from moderngl_window.resources.tracks import tracks  # noqa
from moderngl_window.resources.scenes import scenes  # noqa
from moderngl_window.resources.data import data  # noqa


def register_dir(path: Union[Path, str]) -> None:
    """Adds a resource directory for all resource types

    Args:
        path (Union[Path, str]): Directory path
    """
    register_data_dir(path)
    register_program_dir(path)
    register_scene_dir(path)
    register_texture_dir(path)


def register_program_dir(path: Union[Path, str]) -> None:
    """Adds a resource directory specifically for programs

    Args:
        path (Union[Path, str]): Directory path
    """
    _append_unique_path(path, settings.PROGRAM_DIRS)


def register_texture_dir(path: Union[Path, str]) -> None:
    """Adds a resource directory specifically for textures

    Args:
        path (Union[Path, str]): Directory path
    """
    _append_unique_path(path, settings.TEXTURE_DIRS)


def register_scene_dir(path: Union[Path, str]) -> None:
    """Adds a resource directory specifically for scenes

    Args:
        path (Union[Path, str]): Directory path
    """
    _append_unique_path(path, settings.SCENE_DIRS)


def register_data_dir(path: Union[Path, str]) -> None:
    """Adds a resource directory specifically for data files

    Args:
        path (Union[Path, str]): Directory path
    """
    _append_unique_path(path, settings.DATA_DIRS)


def _append_unique_path(path: Union[Path, str], dest: list):
    path = Path(path)
    if not path.is_absolute():
        raise ImproperlyConfigured("Search path must be absolute: {}".format(path))

    if not path.is_dir():
        raise ImproperlyConfigured("Search path is not a directory: {}".format(path))

    if not path.exists():
        raise ImproperlyConfigured("Search path do not exist: {}".format(path))

    for resource_path in dest:
        if Path(resource_path).samefile(path):
            break
    else:
        dest.append(Path(path).absolute())


@contextmanager
def temporary_dirs(dirs: Union[Path, str]):
    """Temporarily changes all resource directories

    Example::

        with temporary_dirs([path1, path2, path3]):
            # Load some resource here

    Args:
        dirs (Union[Path,str]) list of paths to use
    """
    data_dirs = settings.DATA_DIRS
    program_dirs = settings.PROGRAM_DIRS
    scene_dirs = settings.SCENE_DIRS
    textures_dirs = settings.TEXTURE_DIRS

    settings.DATA_DIRS = dirs
    settings.PROGRAM_DIRS = dirs
    settings.SCENE_DIRS = dirs
    settings.TEXTURE_DIRS = dirs

    try:
        yield dirs
    finally:
        settings.DATA_DIRS = data_dirs
        settings.PROGRAM_DIRS = program_dirs
        settings.SCENE_DIRS = scene_dirs
        settings.TEXTURE_DIRS = textures_dirs



"""
6 cubes with drag and drop texture loading.

For each box, locate an image in your File Manager, and drag and drop onto the box.

Currently only working with the Pyglet backend.
"""
from pathlib import Path
import os

import moderngl
import moderngl_window

from pyrr import Matrix44


class Cubes(moderngl_window.WindowConfig):
    title = "Cubes"
    resizable = True
    aspect_ratio = None
    resource_dir = Path(__file__).parent.resolve() / 'resources'

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # Load the 6 different boxes with different vertex formats
        self.box_top_left = self.load_scene('scenes/box/box-T2F_V3F.obj')
        self.box_top_middle = self.load_scene('scenes/box/box-T2F_V3F.obj')
        self.box_top_right = self.load_scene('scenes/box/box-T2F_V3F.obj')
        self.box_bottom_left = self.load_scene('scenes/box/box-T2F_V3F.obj')
        self.box_bottom_middle = self.load_scene('scenes/box/box-T2F_V3F.obj')
        self.box_bottom_right = self.load_scene('scenes/box/box-T2F_V3F.obj')

        self.resize(*self.wnd.size)

    def render(self, time, frame_time):
        self.ctx.enable_only(moderngl.DEPTH_TEST | moderngl.CULL_FACE)
        rot = Matrix44.from_eulers((time, time/2, time/3))

        # Box top left
        view = Matrix44.from_translation((-5, 2, -10), dtype='f4')
        self.box_top_left.draw(self.projection, view * rot)

        # Box top middle
        view = Matrix44.from_translation((0, 2, -10), dtype='f4')
        self.box_top_middle.draw(self.projection, view * rot)

        # Box top right
        view = Matrix44.from_translation((5, 2, -10), dtype='f4')
        self.box_top_right.draw(self.projection, view * rot)

        # Box bottom left
        view = Matrix44.from_translation((-5, -2, -10), dtype='f4')
        self.box_bottom_left.draw(self.projection, view * rot)

        # Box bottom middle
        view = Matrix44.from_translation((0, -2, -10), dtype='f4')
        self.box_bottom_middle.draw(self.projection, view * rot)

        # Box bottom right
        view = Matrix44.from_translation((5, -2, -10), dtype='f4')
        self.box_bottom_right.draw(self.projection, view * rot)

    def resize(self, width, height):
        self.ctx.viewport = 0, 0, width, height
        self.projection = Matrix44.perspective_projection(45, width / height, 1, 50, dtype='f4')

    def _load_texture(self, path):
        tex = self.load_texture_2d(os.path.relpath(path, self.resource_dir))
        print(type(tex))
        return tex

    def files_dropped_event(self, x, y, paths):
        if x < self.wnd._window.width * 0.33:
            if y < self.wnd._window.height * 0.5:
                # Modify top left box
                self.box_top_left.materials[0].mat_texture.texture = \
                    self._load_texture(paths[0])
            else:
                # Modify bottom left box
                self.box_bottom_left.materials[0].mat_texture.texture = \
                    self._load_texture(paths[0])
        elif x < self.wnd._window.width * 0.66:
            if y < self.wnd._window.height * 0.5:
                # Modify top middle box
                self.box_top_middle.materials[0].mat_texture.texture = \
                    self._load_texture(paths[0])
            else:
                # Modify bottom middle box
                self.box_bottom_middle.materials[0].mat_texture.texture = \
                    self._load_texture(paths[0])
        else:
            if y < self.wnd._window.height * 0.5:
                # Modify top right box
                self.box_top_right.materials[0].mat_texture.texture = \
                    self._load_texture(paths[0])
            else:
                # Modify bottom right box
                self.box_bottom_right.materials[0].mat_texture.texture = \
                    self._load_texture(paths[0])
        print(paths)

if __name__ == '__main__':
    Cubes.run()



"""
Base finders
"""
import functools
import logging

from collections import namedtuple
from pathlib import Path

from moderngl_window.conf import settings
from moderngl_window.exceptions import ImproperlyConfigured
from moderngl_window.utils.module_loading import import_string

FinderEntry = namedtuple("FinderEntry", ["path", "abspath", "exists"])

logger = logging.getLogger(__name__)


class BaseFilesystemFinder:
    """Base class for searching filesystem directories"""

    settings_attr = None
    """str: Name of the attribute in :py:class:`~moderngl_window.conf.Settings`
    containing a list of paths the finder should search in.
    """

    def __init__(self):
        """Initialize finder class by looking up the paths referenced in ``settings_attr``.
        """
        if not hasattr(settings, self.settings_attr):
            raise ImproperlyConfigured(
                "Settings doesn't define {}. "
                "This is required when using a FileSystemFinder.".format(
                    self.settings_attr
                )
            )
        self.paths = getattr(settings, self.settings_attr)

    def find(self, path: Path) -> Path:
        """Finds a file in the configured paths returning its absolute path.

        Args:
            path (pathlib.Path): The path to find
        Returns:
            The absolute path to the file or None if not found
        """
        # Update paths from settings to make them editable runtime
        if getattr(self, "settings_attr", None):
            self.paths = getattr(settings, self.settings_attr)

        if not isinstance(path, Path):
            raise ValueError(
                "FilesystemFinders only take Path instances, not {}".format(type(path))
            )

        logger.debug("find %s", path)

        # Ignore absolute paths so other finder types can pick them up.
        if path.is_absolute():
            logger.debug("Ignoring absolute path: %s", path)
            return None

        logger.debug("paths: %s", self.paths)

        for search_path in self.paths:
            search_path = Path(search_path)

            # Keep ensuring all search paths are absolute
            if not search_path.is_absolute():
                raise ImproperlyConfigured(
                    "Search search path '{}' is not an absolute path"
                )

            abspath = search_path / path
            logger.debug("abspath %s", abspath)

            if abspath.exists():
                logger.debug("found %s", abspath)
                return abspath

        return None


@functools.lru_cache(maxsize=None)
def get_finder(import_path: str):
    """
    Get a finder class from an import path.
    This function uses an lru cache.

    Args:
        import_path: string representing an import path
    Return:
        An instance of the finder
    Raises:
        ImproperlyConfigured is the finder is not found
    """
    Finder = import_string(import_path)
    if not issubclass(Finder, BaseFilesystemFinder):
        raise ImproperlyConfigured(
            "Finder {} is not a subclass of .finders.FileSystemFinder".format(
                import_path
            )
        )

    return Finder()



from unittest import TestCase
import numpy as np
from pyrr import Vector3

from moderngl_window.scene import Camera, KeyboardCamera
from moderngl_window.scene import camera as cam
from moderngl_window.opengl.projection import Projection3D
from moderngl_window.context.base.keys import BaseKeys, KeyModifiers


class CameraTest(TestCase):
    """Simple camera testing"""

    def test_camera(self):
        camera = Camera(fov=60, aspect_ratio=1.0, near=1.0, far=100.0)
        self.assertIsInstance(camera.projection, Projection3D)
        self.assertIsInstance(camera.matrix, np.ndarray)

        camera.look_at(vec=Vector3((1, 2, 3)))
        camera.look_at(pos=(4, 5, 6))
        camera.set_position(1, 1, 1)

    def test_keyboardcamera(self):
        camera = KeyboardCamera(BaseKeys, fov=60, aspect_ratio=1.0, near=1.0, far=100.0)
        camera.mouse_sensitivity = 10.0
        camera.velocity = 10.0
        self.assertEqual(camera.mouse_sensitivity, 10.0)
        self.assertEqual(camera.velocity, 10.0)
        self.assertIsInstance(camera.projection, Projection3D)
        self.assertIsInstance(camera.matrix, np.ndarray)

        camera.key_input(BaseKeys.UP, BaseKeys.ACTION_PRESS, KeyModifiers)

        camera.move_left(True)
        camera.move_right(True)
        camera.move_forward(True)
        camera.move_backward(True)
        camera.move_up(True)
        camera.move_down(True)

        camera.move_state(cam.FORWARD, True)
        camera.rot_state(-17, 4)



"""
Documentation testing
Inspired by:
https://github.com/cprogrammer1994/ModernGL/blob/master/tests/test_documentation.py
by Szabolcs Dombi
This version is simplified:
* Only test if the attribute or method is present in the class. Function parameters are not inspected.
* Include ignore pattern in the implemented set
"""
import os
import re
import sys
import types
import unittest
from unittest.mock import MagicMock

from moderngl_window.utils import module_loading

# Mock modules
class Mock(MagicMock):
    @classmethod
    def __getattr__(cls, name):
        return MagicMock()


MOCK_MODULES = [
    'glfw',
    'sdl2',
    'sdl2.ext',
    'sdl2.video',
    'pyglet',
    'pyglet.window',
    'PyQt5',
    'PyQt5.QtCore',
    'QtCore',
    'QtOpenGL',
    'QtWidgets',
    'PySide2',
    'PySide2.QtCore',
]

sys.modules.update((mod_name, Mock()) for mod_name in MOCK_MODULES)


class TestCase(unittest.TestCase):
    """
    Test reference docs
    """
    def validate(self, filename, module, classname=None, ignore=None):
        """
        Finds all automethod and autoattribute statements in an rst file
        comparing them to the attributes found in the actual class
        """
        if ignore is None:
            ignore = []

        with open(os.path.normpath(os.path.join('docs', 'reference', filename))) as f:
            docs = f.read()

        module = module_loading.import_module(module)

        # Inspect class
        if classname:
            methods = re.findall(r'^\.\. automethod:: ([^\(\n]+)', docs, flags=re.M)
            attributes = re.findall(r'^\.\. autoattribute:: ([^\n]+)', docs, flags=re.M)

            documented = set(filter(lambda x: x.startswith(classname), [a for a in methods] + attributes))
            implemented = set(classname + '.' + x for x in dir(getattr(module, classname))
                              if not x.startswith('_') or x == '__init__')
            ignored = set(classname + '.' + x for x in ignore)
        # Inspect module
        else:
            # Only inspect functions for now
            functions = re.findall(r'^\.\. autofunction:: ([^\(\n]+)', docs, flags=re.M)
            documented = set(functions)
            ignored = set(ignore)
            implemented = set(func for func in dir(module) if isinstance(getattr(module, func), types.FunctionType))

        self.assertSetEqual(implemented - documented - ignored, set(), msg='Implemented but not Documented')
        self.assertSetEqual(documented - implemented - ignored, set(), msg='Documented but not Implemented')

    def test_moderngl_window(self):
        self.validate(
            'moderngl_window.rst',
            'moderngl_window',
            ignore=['valid_window_size', 'valid_window_size_multiplier', 'import_string', 'valid_bool', "QWERTY", "AZERTY"],
        )

    def test_settings(self):
        self.validate('settings.conf.settings.rst', 'moderngl_window.conf', 'Settings', [])

    # --- context ---

    def test_context_base_window(self):
        self.validate('context/basewindow.rst', 'moderngl_window.context.base.window', 'BaseWindow')

    def test_context_glfw_window(self):
        self.validate('context/glfw.window.rst', 'moderngl_window.context.glfw.window', 'Window')

    def test_context_headless_window(self):
        self.validate('context/headless.window.rst', 'moderngl_window.context.headless.window', 'Window')

    def test_context_pyglet_window(self):
        self.validate('context/pyglet.window.rst', 'moderngl_window.context.pyglet.window', 'Window')

    def test_context_pyqt5_window(self):
        self.validate('context/pyqt5.window.rst', 'moderngl_window.context.pyqt5.window', 'Window')

    # @unittest.skipIf(sys.version_info >= (3, 8, 0), reason="pyside2 not supported in py38 yet")
    def test_context_pyside2_window(self):
        self.validate('context/pyside2.window.rst', 'moderngl_window.context.pyside2.window', 'Window')

    def test_context_sdl2_window(self):
        self.validate('context/sdl2.window.rst', 'moderngl_window.context.sdl2.window', 'Window')

    # --- geometry ---

    def test_geometry(self):
        self.validate('geometry.rst', 'moderngl_window.geometry')

    # --- Loaders ---

    def test_loaders_base(self):
        self.validate('loaders/base.rst', 'moderngl_window.loaders.base', 'BaseLoader')

    # --- Loaders : Texture ---

    def test_loaders_t2d(self):
        self.validate('loaders/t2d.rst', 'moderngl_window.loaders.texture.t2d', 'Loader')

    def test_loaders_array(self):
        self.validate('loaders/array.rst', 'moderngl_window.loaders.texture.array', 'Loader')

    # --- Loaders : Scene ---

    def test_loaders_wavefront(self):
        self.validate('loaders/wavefront.rst', 'moderngl_window.loaders.scene.wavefront', 'Loader')

    def test_loaders_gltf(self):
        self.validate('loaders/gltf2.rst', 'moderngl_window.loaders.scene.gltf2', 'Loader')

    def test_loaders_stl(self):
        self.validate('loaders/wavefront.rst', 'moderngl_window.loaders.scene.stl', 'Loader')

    # --- Loaders : Program ---

    def test_loader_single(self):
        self.validate('loaders/single.rst', 'moderngl_window.loaders.program.single', 'Loader')

    def test_loader_separate(self):
        self.validate('loaders/separate.rst', 'moderngl_window.loaders.program.separate', 'Loader')

    # --- Loaders : Data ---

    def test_loader_text(self):
        self.validate('loaders/text.rst', 'moderngl_window.loaders.data.text', 'Loader')

    def test_loader_json(self):
        self.validate('loaders/json.rst', 'moderngl_window.loaders.data.json', 'Loader')

    def test_loader_binary(self):
        self.validate('loaders/binary.rst', 'moderngl_window.loaders.data.binary', 'Loader')

    # --- Meta ---

    def test_meta_base(self):
        self.validate('meta/base.rst', 'moderngl_window.meta.base', 'ResourceDescription')

    def test_meta_texture(self):
        self.validate('meta/texture.rst', 'moderngl_window.meta.texture', 'TextureDescription')

    def test_meta_program(self):
        self.validate('meta/program.rst', 'moderngl_window.meta.program', 'ProgramDescription')

    def test_meta_scene(self):
        self.validate('meta/scene.rst', 'moderngl_window.meta.scene', 'SceneDescription')

    def test_meta_data(self):
        self.validate('meta/data.rst', 'moderngl_window.meta.data', 'DataDescription')

    # --- Finders ---

    def test_finders_base(self):
        self.validate('finders/base.rst', 'moderngl_window.finders.base', 'BaseFilesystemFinder')

    def test_finders_texture(self):
        self.validate('finders/texture.rst', 'moderngl_window.finders.texture', 'FilesystemFinder')

    def test_finders_program(self):
        self.validate('finders/program.rst', 'moderngl_window.finders.program', 'FilesystemFinder')

    def test_finders_scene(self):
        self.validate('finders/scene.rst', 'moderngl_window.finders.scene', 'FilesystemFinder')

    def test_finders_data(self):
        self.validate('finders/data.rst', 'moderngl_window.finders.data', 'FilesystemFinder')

    # --- opengl ---

    def test_opengl_projection3d(self):
        self.validate('opengl/projection.rst', 'moderngl_window.opengl.projection', 'Projection3D')

    def test_opengl_vao(self):
        self.validate('opengl/vao.rst', 'moderngl_window.opengl.vao', 'VAO')

    # --- resources ---

    def test_resources_base(self):
        self.validate('resources/base.rst', 'moderngl_window.resources.base', 'BaseRegistry')

    def test_resources_data(self):
        self.validate('resources/data.rst', 'moderngl_window.resources.data', 'DataFiles')

    def test_resources_textures(self):
        self.validate('resources/textures.rst', 'moderngl_window.resources.textures', 'Textures')

    def test_resources_programs(self):
        self.validate('resources/programs.rst', 'moderngl_window.resources.programs', 'Programs')

    def test_resources_scenes(self):
        self.validate('resources/scenes.rst', 'moderngl_window.resources.scenes', 'Scenes')

    # --- timers ---

    def test_timers_base(self):
        self.validate('timers/base.rst', 'moderngl_window.timers.base', 'BaseTimer')

    def test_timers_clock(self):
        self.validate('timers/clock.rst', 'moderngl_window.timers.clock', 'Timer')

    # -- Scene ---

    def test_scene_camera(self):
        self.validate('scene/camera.rst', 'moderngl_window.scene', 'Camera')

    def test_scene_keyboardcamera(self):
        self.validate('scene/keyboardcamera.rst', 'moderngl_window.scene', 'KeyboardCamera')

    def test_scene_scene(self):
        self.validate('scene/scene.rst', 'moderngl_window.scene', 'Scene')

    def test_scene_node(self):
        self.validate('scene/node.rst', 'moderngl_window.scene', 'Node')

    def test_scene_mesh(self):
        self.validate('scene/mesh.rst', 'moderngl_window.scene', 'Mesh')

    def test_scene_material(self):
        self.validate('scene/material.rst', 'moderngl_window.scene', 'Material')

    def test_scene_material_texture(self):
        self.validate('scene/materialtexture.rst', 'moderngl_window.scene', 'MaterialTexture')

    def test_scene_meshprogram(self):
        self.validate('scene/meshprogram.rst', 'moderngl_window.scene', 'MeshProgram')



import numpy

import moderngl
from moderngl_window.opengl.vao import VAO
from moderngl_window.geometry.attributes import AttributeNames


def quad_fs(attr_names=AttributeNames, normals=True, uvs=True, name=None) -> VAO:
    """
    Creates a screen aligned quad using two triangles with normals and texture coordinates.

    Keyword Args:
        attr_names (AttributeNames): Attrib name config
        normals (bool): Include normals in VAO
        uvs (bool): Include texture coordinates in VAO
        name (str): Optional name for the VAO
    Returns:
        A :py:class:`~moderngl_window.opengl.vao.VAO` instance.
    """
    return quad_2d(
        size=(2.0, 2.0), normals=normals, uvs=uvs, attr_names=attr_names, name=name,
    )


def quad_2d(
    size=(1.0, 1.0),
    pos=(0.0, 0.0),
    normals=True,
    uvs=True,
    attr_names=AttributeNames,
    name=None,
) -> VAO:
    """
    Creates a 2D quad VAO using 2 triangles with normals and texture coordinates.

    Keyword Args:
        size (tuple): width and height
        pos (float): Center position x and y
        normals (bool): Include normals in VAO
        uvs (bool): Include texture coordinates in VAO
        attr_names (AttributeNames): Attrib name config
        name (str): Optional name for the VAO
    Returns:
        A :py:class:`~moderngl_window.opengl.vao.VAO` instance.
    """
    width, height = size
    xpos, ypos = pos

    # fmt: off
    pos_data = numpy.array([
        xpos - width / 2.0, ypos + height / 2.0, 0.0,
        xpos - width / 2.0, ypos - height / 2.0, 0.0,
        xpos + width / 2.0, ypos - height / 2.0, 0.0,
        xpos - width / 2.0, ypos + height / 2.0, 0.0,
        xpos + width / 2.0, ypos - height / 2.0, 0.0,
        xpos + width / 2.0, ypos + height / 2.0, 0.0,
    ], dtype=numpy.float32)

    normal_data = numpy.array([
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
        0.0, 0.0, 1.0,
    ], dtype=numpy.float32)

    uv_data = numpy.array([
        0.0, 1.0,
        0.0, 0.0,
        1.0, 0.0,
        0.0, 1.0,
        1.0, 0.0,
        1.0, 1.0,
    ], dtype=numpy.float32)
    # fmt: on

    vao = VAO(name or "geometry:quad", mode=moderngl.TRIANGLES)
    vao.buffer(pos_data, "3f", [attr_names.POSITION])
    if normals:
        vao.buffer(normal_data, "3f", [attr_names.NORMAL])
    if uvs:
        vao.buffer(uv_data, "2f", [attr_names.TEXCOORD_0])

    return vao



import math

import numpy

import moderngl as mlg
from moderngl_window.opengl.vao import VAO
from moderngl_window.geometry import AttributeNames


def sphere(
    radius=0.5,
    sectors=32,
    rings=16,
    normals=True,
    uvs=True,
    name: str = None,
    attr_names=AttributeNames,
) -> VAO:
    """Creates a sphere.

    Keyword Args:
        radius (float): Radius or the sphere
        rings (int): number or horizontal rings
        sectors (int): number of vertical segments
        normals (bool): Include normals in the VAO
        uvs (bool): Include texture coordinates in the VAO
        name (str): An optional name for the VAO
        attr_names (AttributeNames): Attribute names
    Returns:
        A :py:class:`VAO` instance
    """
    R = 1.0 / (rings - 1)
    S = 1.0 / (sectors - 1)

    vertices = [0] * (rings * sectors * 3)
    normals = [0] * (rings * sectors * 3)
    uvs = [0] * (rings * sectors * 2)

    v, n, t = 0, 0, 0
    for r in range(rings):
        for s in range(sectors):
            y = math.sin(-math.pi / 2 + math.pi * r * R)
            x = math.cos(2 * math.pi * s * S) * math.sin(math.pi * r * R)
            z = math.sin(2 * math.pi * s * S) * math.sin(math.pi * r * R)

            uvs[t] = s * S
            uvs[t + 1] = r * R

            vertices[v] = x * radius
            vertices[v + 1] = y * radius
            vertices[v + 2] = z * radius

            normals[n] = x
            normals[n + 1] = y
            normals[n + 2] = z

            t += 2
            v += 3
            n += 3

    indices = [0] * rings * sectors * 6
    i = 0
    for r in range(rings - 1):
        for s in range(sectors - 1):
            indices[i] = r * sectors + s
            indices[i + 1] = (r + 1) * sectors + (s + 1)
            indices[i + 2] = r * sectors + (s + 1)

            indices[i + 3] = r * sectors + s
            indices[i + 4] = (r + 1) * sectors + s
            indices[i + 5] = (r + 1) * sectors + (s + 1)
            i += 6

    vao = VAO(name or "sphere", mode=mlg.TRIANGLES)

    vbo_vertices = numpy.array(vertices, dtype=numpy.float32)
    vao.buffer(vbo_vertices, "3f", [attr_names.POSITION])

    if normals:
        vbo_normals = numpy.array(normals, dtype=numpy.float32)
        vao.buffer(vbo_normals, "3f", [attr_names.NORMAL])

    if uvs:
        vbo_uvs = numpy.array(uvs, dtype=numpy.float32)
        vao.buffer(vbo_uvs, "2f", [attr_names.TEXCOORD_0])

    vbo_elements = numpy.array(indices, dtype=numpy.uint32)
    vao.index_buffer(vbo_elements, index_element_size=4)

    return vao



import os
import sys

from types import ModuleType
from unittest import TestCase

from moderngl_window.conf import Settings, SETTINGS_ENV_VAR
from moderngl_window.exceptions import ImproperlyConfigured


class SettingsTests(TestCase):
    """Test settings system"""
    # Non-standard window settings
    window_setting = {
        'class': 'moderngl_window.context.headless.Window',
        'name': 'ModernGL Headless Test',
        'gl_version': (3, 3),
    }

    def test_default(self):
        """Initialize default settings"""
        settings = Settings()
        # Attempt to access default properties
        settings.WINDOW
        settings.PROGRAM_FINDERS
        settings.PROGRAM_DIRS
        settings.PROGRAM_LOADERS

    def test_apply_settings_from_env(self):
        """Config values from module"""
        # Create and inject a fake module in python
        module_name = "moderngl_window_settings_module"
        module = ModuleType(module_name)
        setattr(module, 'WINDOW', self.window_setting)
        sys.modules[module_name] = module

        # Attempt to import it
        os.environ[SETTINGS_ENV_VAR] = module_name
        settings = Settings()
        settings.apply_settings_from_env()
        self.assertEqual(settings.WINDOW, self.window_setting)

    def test_import_nonexistent_module(self):
        """Import settings module that do no exist"""
        os.environ[SETTINGS_ENV_VAR] = 'this.module.does.not.exist'
        with self.assertRaises(ImproperlyConfigured):
            Settings().apply_settings_from_env()

    def test_apply_from_dict(self):
        """Supply config values as dict"""
        settings = Settings()
        settings.apply_from_dict({'WINDOW': self.window_setting})
        self.assertEqual(settings.WINDOW, self.window_setting)

    def test_apply_from_cls(self):
        """Supply config values using cls namespace"""
        class MyConfig:
            WINDOW=self.window_setting

        settings = Settings()
        settings.apply_from_cls(MyConfig)
        self.assertEqual(settings.WINDOW, self.window_setting)

    def test_apply_from_generator(self):
        """Test generators specifically"""
        def gen():
            yield 'WINDOW', self.window_setting
            yield 'SOMETHING', 1

        settings = Settings()
        settings.apply_from_iterable(gen())
        self.assertEqual(settings.WINDOW, self.window_setting)
        self.assertEqual(settings.SOMETHING, 1)

    def test_apply_from_iterator_error(self):
        """Ensure ValueError is rasied if not iterable"""
        with self.assertRaises(ValueError):
            settings = Settings()
            settings.apply_from_iterable(1337)

    def test_repr(self):
        """Ensure string represenation is somewhat reasonable"""
        value = str(Settings())
        self.assertIsInstance(value, str)
        self.assertGreater(len(value), 100)



"""
Scene Registry
"""
from moderngl_window.resources.base import BaseRegistry
from moderngl_window.scene import Scene
from moderngl_window.meta import SceneDescription


class Scenes(BaseRegistry):
    """Handles scene loading"""

    settings_attr = "SCENE_LOADERS"

    def load(self, meta: SceneDescription) -> Scene:
        """Load a scene with the configured loaders.

        Args:
            meta (:py:class:`~moderngl_window.meta.scene.SceneDescription`): The resource description
        Returns:
            :py:class:`~moderngl_window.scene.Scene`: The loaded scene
        """
        return super().load(meta)


scenes = Scenes()



from pathlib import Path

import moderngl
from headless import HeadlessTestCase
from moderngl_window.context.headless import Keys
from moderngl_window import resources
from moderngl_window.meta import ProgramDescription
from moderngl_window import geometry

resources.register_dir((Path(__file__).parent / 'fixtures' / 'resources').resolve())


class HeadlessWindowTestCase(HeadlessTestCase):
    """Basic property and method testing"""
    window_size = (16, 16)
    aspect_ratio = 1.0

    def test_properties(self):
        self.assertEqual(self.window.keys, Keys)
        self.assertEqual(self.window.title, 'ModernGL Window')
        self.assertIsInstance(self.window.ctx, moderngl.Context)
        self.assertIsInstance(self.window.fbo, moderngl.Framebuffer)
        self.assertEqual(self.window.gl_version, (4, 1))
        self.assertEqual(self.window.width, self.window_size[0])
        self.assertEqual(self.window.height, self.window_size[1])
        self.assertEqual(self.window.size, self.window_size)
        self.assertEqual(self.window.buffer_size, self.window_size)
        self.assertEqual(self.window.viewport, (0, 0, self.window_size[0], self.window_size[1]))
        self.assertIsInstance(self.window.frames, int)
        self.assertEqual(self.window.resizable, False)
        self.assertEqual(self.window.fullscreen, False)
        self.assertEqual(self.window.config, None)
        self.assertEqual(self.window.vsync, False)
        self.assertEqual(self.window.aspect_ratio, 1.0)
        self.assertEqual(self.window.samples, 0)
        self.assertEqual(self.window.cursor, False)
        self.assertIsNotNone(self.window.modifiers)
        self.assertFalse(self.window.is_closing)

    def test_swap_buffers(self):
        """Ensure frame counter increases"""
        frame = self.window.frames
        self.window.swap_buffers()
        self.assertEqual(self.window.frames, frame + 1)

    def test_functions(self):
        """Ensure the basic functions do not cause crashes"""
        self.window.swap_buffers()
        self.window.clear()
        self.window.render(0, 0)
        self.window.print_context_info()
        self.assertFalse(self.window.is_key_pressed(self.window.keys.ESCAPE))

    def test_resize(self):
        """Resize should do nothing"""
        self.window.resize(16, 16)

    def test_viewport(self):
        """Set default viewport"""
        self.window.set_default_viewport()
        self.assertEqual(self.window.viewport, (0, 0, self.window_size[0], self.window_size[1]))

    def test_render(self):
        """Render something simple to the framebuffer"""
        self.window.use()
        self.window.clear()
        prog = resources.programs.load(ProgramDescription(path="programs/white.glsl"))
        quad = geometry.quad_fs()
        quad.render(prog)

        # Ensure all fragments (rgba) values are white
        data = self.window.fbo.read(components=4)
        self.assertEqual(data, b'\xff' * (self.window_size[0] * self.window_size[1] * 4))



from .base import BaseVideoCapture
import subprocess
import moderngl


class FFmpegCapture(BaseVideoCapture):
    """
        ``FFmpegCapture`` it's an utility class to capture runtime render
        and save it as video.

        Args:


        Example:

        .. code:: python

            import moderngl_window
            from moderngl_window.capture.ffmpeg import FFmpegCapture

            class CaptureTest(modenrgl_window.WindowConfig):

                def __init__(self, **kwargs):
                    super().__init__(**kwargs)
                    # do other initialization

                    # define VideoCapture class
                    self.cap = FFmpegCapture(source=self.wnd.fbo)

                    # start recording
                    self.cap.start_capture(
                        filename="video.mp4",
                        framerate=30
                    )

                def render(self, time, frametime):
                    # do other render stuff

                    # call record function after
                    self.cap.save()

                def close(self):
                    # if realease func is not called during
                    # runtime. make sure to do it on the closing event
                    self.cap.release()

    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self._ffmpeg = None

    def _start_func(self) -> bool:
        """
            choose the right pixel format based on the number of components
            and start a ffmper pipe with a subprocess.
        """
        pix_fmt = 'rgb24'   # 3 component, 1 byte per color -> 24 bit

        # for the framebuffer is easier because i can read 3 component even if
        # the color attachment has less components
        if isinstance(self._source, moderngl.Texture) and self._components == 4:
            pix_fmt = 'rgba'  # 4 component , 1 byte per color -> 32 bit

        command = [
            'ffmpeg',
            '-hide_banner',
            '-loglevel', 'error', '-stats',  # less verbose, only stats of recording
            '-y',  # (optional) overwrite output file if it exists
            '-f', 'rawvideo',
            '-vcodec', 'rawvideo',
            '-s', f'{self._width}x{self._height}',  # size of one frame
            '-pix_fmt', pix_fmt,
            '-r', f'{self._framerate}',  # frames per second
            '-i', '-',  # The imput comes from a pipe
            '-vf', 'vflip',
            '-an',  # Tells FFMPEG not to expect any audio
            self._filename,
        ]

        # ffmpeg binary need to be on the PATH.
        try:
            self._ffmpeg = subprocess.Popen(
                command,
                stdin=subprocess.PIPE,
                bufsize=0
            )
        except FileNotFoundError:
            print("ffmpeg command not found. Be sure to add it to PATH")
            return

        return True

    def _release_func(self):
        """
        Safely release the capture
        """
        self._ffmpeg.stdin.close()
        _ = self._ffmpeg.wait()

    def _dump_frame(self, frame):
        """
        write the frame data in to the ffmpeg pipe
        """
        self._ffmpeg.stdin.write(frame)
